# -*- coding: utf-8 -*-
#
# @fileoverview Copyright (c) 2019-2021, Stefano Gualandi,
#                via Ferrata, 1, I-27100, Pavia, Italy
#
#  @author stefano.gualandi@gmail.com (Stefano Gualandi)
#

cdef extern from "OT1D.hpp":
	cdef double OT1Da0(int, double*, double*, bool, int)
	cdef double OT1Db0(int, double*, double*, bool, int)
	cdef double OT1Da(int, int, double*, double*, bool, int)
	cdef double OT1Db(int, int, double*, double*, bool, int)
	cdef double OT1Dc(int, int, double*, double*, double*, double*, bool, int)
	cdef double OT1Dd(int, int, double*, double*, double*, double*, bool, int)
	cdef void _parasort(int, double*, int)
	cdef void _parasort_mu(int, double*, double*, int)


# distutils: language=c++

# -*- coding: utf-8 -*-
#
# @fileoverview Copyright (c) 2019-2021, Stefano Gualandi,
#                via Ferrata, 1, I-27100, Pavia, Italy
#
#  @author stefano.gualandi@gmail.com (Stefano Gualandi)
#

import numpy as np

from OT1D cimport OT1Da0 as OT1Da0
from OT1D cimport OT1Db0 as OT1Db0
from OT1D cimport OT1Da as OT1Da
from OT1D cimport OT1Db as OT1Db
from OT1D cimport OT1Dc as OT1Dc
from OT1D cimport OT1Dd as OT1Dd
from OT1D cimport _parasort as _parasort
from OT1D cimport _parasort_mu as _parasort_mu

# One function for optimal transport
def OT1D(X, Y, A=None, B=None, p=1, sorting=True, threads=8):
    # Wassertein of order 1
    if p == 1:
        if A is None and B is None:
            return OT_1Da(X, Y, sorting, threads)
        else:
            return OT_1Dc(X, Y, A, B, sorting, threads)

    # Wassertein of order 2
    if p == 2:
        if A is None and B is None:
            return OT_1Db(X, Y, sorting, threads)
        else:
            return OT_1Dd(X, Y, A, B, sorting, threads)


# Support for fast sorting
def parasort(X, A=None, threads=8):
    if not X.flags['C_CONTIGUOUS'] or type(X) == list:
        X = np.ascontiguousarray(X, dtype=float) 
    cdef double[::1] Xmv = X

    if A is None:
        return _parasort(len(X), &Xmv[0], threads)

    if not A.flags['C_CONTIGUOUS'] or type(A) == list:
        A = np.ascontiguousarray(A, dtype=float) 
    cdef double[::1] Amv = A    

    return _parasort_mu(len(X), &Xmv[0], &Amv[0], threads)


#--------------------------------------------------------------
# FROM C++
#--------------------------------------------------------------
cdef double OT_1Da(X, Y, sorting=True, threads=8):
    m = len(X)
    n = len(Y)
    if not X.flags['C_CONTIGUOUS'] or type(X) == list:
        X = np.ascontiguousarray(X, dtype=float) 
    cdef double[::1] Xmv = X
        
    if not Y.flags['C_CONTIGUOUS'] or type(Y) == list:
        Y = np.ascontiguousarray(Y, dtype=float) 
    cdef double[::1] Ymv = Y
    
    if m == n:
        return OT1Da0(n, &Xmv[0], &Ymv[0], sorting, threads) 
    return OT1Da(m, n, &Xmv[0], &Ymv[0], sorting, threads)


cdef double OT_1Db(X, Y, sorting=True, threads=8):
    m = len(X)
    n = len(Y)
    if not X.flags['C_CONTIGUOUS'] or type(X) == list:
        X = np.ascontiguousarray(X, dtype=float) 
    cdef double[::1] Xmv = X
        
    if not Y.flags['C_CONTIGUOUS'] or type(Y) == list:
        Y = np.ascontiguousarray(Y, dtype=float) 
    cdef double[::1] Ymv = Y

    if m == n:
        return OT1Db0(n, &Xmv[0], &Ymv[0], sorting, threads)     
    return OT1Db(m, n, &Xmv[0], &Ymv[0], sorting, threads)


cdef double OT_1Dc(X, Y, A, B, sorting=True, threads=8):
    m = len(X)
    n = len(Y)
    if not X.flags['C_CONTIGUOUS'] or type(X) == list:
        X = np.ascontiguousarray(X, dtype=float) 
    cdef double[::1] Xmv = X
        
    if not Y.flags['C_CONTIGUOUS'] or type(Y) == list:
        Y = np.ascontiguousarray(Y, dtype=float)
    cdef double[::1] Ymv = Y

    if not A.flags['C_CONTIGUOUS'] or type(A) == list:
        A = np.ascontiguousarray(A, dtype=float) 
    cdef double[::1] Amv = A
        
    if not B.flags['C_CONTIGUOUS'] or type(B) == list:
        B = np.ascontiguousarray(B, dtype=float) 
    cdef double[::1] Bmv = B
    
    return OT1Dc(m, n, &Xmv[0], &Ymv[0], &Amv[0], &Bmv[0], sorting, threads)


cdef double OT_1Dd(X, Y, A, B, sorting=True, threads=8):
    m = len(X)
    n = len(Y)
    if not X.flags['C_CONTIGUOUS'] or type(X) == list:
        X = np.ascontiguousarray(X, dtype=float) 
    cdef double[::1] Xmv = X
        
    if not Y.flags['C_CONTIGUOUS'] or type(Y) == list:
        Y = np.ascontiguousarray(Y, dtype=float)
    cdef double[::1] Ymv = Y

    if not A.flags['C_CONTIGUOUS'] or type(A) == list:
        A = np.ascontiguousarray(A, dtype=float) 
    cdef double[::1] Amv = A
        
    if not B.flags['C_CONTIGUOUS'] or type(B) == list:
        B = np.ascontiguousarray(B, dtype=float) 
    cdef double[::1] Bmv = B
    
    return OT1Dd(m, n, &Xmv[0], &Ymv[0], &Amv[0], &Bmv[0], sorting, threads)




# -*- coding: utf-8 -*-
#
# @fileoverview Copyright (c) 2019-2021, Stefano Gualandi,
#                via Ferrata, 1, I-27100, Pavia, Italy
#
#  @author stefano.gualandi@gmail.com (Stefano Gualandi)
#

import numpy as np
from OT1D import OT1D, parasort

np.random.seed(13)

N = 1000000

# Uniform samples
x = np.random.uniform(1, 2, N)
y = np.random.uniform(0, 1, N)

z = OT1D(x, y, p=2, sorting=True, threads=16)

print('Wasserstein distance of order 2, W2(x,y) =', z)

# Output:
# Wasserstein distance of order 2, W2(x,y) = 1.0002549459050794


# -*- coding: utf-8 -*-
#
# @fileoverview Copyright (c) 2019-2021, Stefano Gualandi,
#                via Ferrata, 1, I-27100, Pavia, Italy
#
#  @author stefano.gualandi@gmail.com (Stefano Gualandi)
#

from setuptools import setup, Extension, Command, find_packages
from Cython.Build import cythonize

import numpy as np
import platform

CC_ARGS = [
    '-Wno-unused-function', '-std=c++11', '-O2', '-ffast-math',
    '-DNDEBUG', '-fno-wrapv', '-march=native'
]

LD_ARGS = ['-O2', '-lm', '-pthread', '-fno-wrapv']

if platform.system() == 'Windows':
    # Compiler options: 
    # https://docs.microsoft.com/en-us/cpp/build/reference/openmp-enable-openmp-2-0-support
    CC_ARGS = ['/std:c++14', '/arch:AVX2', '/O2']
    LD_ARGS = []

if platform.system() == 'Darwin':
    CC_ARGS = [
        '-std=c++11', '-stdlib=libc++', '-O2', '-ffast-math', '-DNDEBUG'
    ]
    LD_ARGS = ['-O2', '-lm', '-pthread', '-fno-wrapv']


extensions = Extension(name="OT1D",
                       sources=["OT1D.pyx"],
                       include_dirs=['./'],
                       extra_compile_args=CC_ARGS,
                       extra_link_args=LD_ARGS)

with open('README.md', encoding="utf-8") as f:
    long_descr = f.read()

setup(
    name='OT1D',
    version='0.3.4',
    packages=find_packages(),
    description='Discrete Optimal Transport in 1D by Linear Programming',
    author='Stefano Gualandi',
    author_email='stefano.gualandi@gmail.com',
    url='https://github.com/stegua/ot1d',
    platforms=['linux', 'macosx', 'windows'],
    setup_requires=['numpy', 'cython'],
    install_requires=['numpy', 'cython'],
    include_dirs=np.get_include(),
    long_description=long_descr,
    long_description_content_type='text/markdown',
    ext_modules = cythonize(extensions, 
        compiler_directives={'language_level': "3"})
)



# -*- coding: utf-8 -*-
"""
Created on Mon Apr  6 10:57:37 2020

@author: Gualandi
"""

import numpy as np
from time import perf_counter

from scipy.stats import wasserstein_distance as W1
import ot


from OT1D import OT1D, parasort


def Test1(p, q, threads=8):
    sorting=True
    print('\n--------------- TEST 1: Unsorted input --------------------')

    print('For OT1D using', threads,'threads\n')

    print("Testing W1, samples of deltas, n=m")
    t0 = perf_counter()
    z = W1(p, q)
    print("Scipy: z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = ot.lp.wasserstein_1d(p, q, p=1)
    print("POT  : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = OT1D(p, q, p=1, sorting=sorting, threads=threads)
    print("SG   : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    print("\nTesting W2, samples of deltas, n=m")
    t0 = perf_counter()
    z = ot.lp.wasserstein_1d(p, q, p=2)
    print("POT  : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = OT1D(p, q, p=2, sorting=sorting, threads=threads)
    print("SG   : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))


    print("\nTesting W1, samples with weights")
    t0 = perf_counter()
    z = W1(p, q, a, b)
    print("Scipy: z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = ot.lp.wasserstein_1d(p, q, a, b, p=1)
    print("POT  : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = OT1D(p, q, a, b, p=1, sorting=sorting, threads=threads)
    print("SG   : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    print("\nTesting W2, samples with weights")
    t0 = perf_counter()
    z = ot.lp.wasserstein_1d(p, q, a, b, p=2)
    print("POT  : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = OT1D(p, q, a, b, p=2, sorting=sorting, threads=threads)
    print("SG   : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))
        
    
    
def Test2(p, q, a, b, threads=8):
    print('\n--------------- TEST 2: Sorted input --------------------')

    print('For OT1D using', threads,'threads\n')

    sorting=False

    t0 = perf_counter()
    # Sort input data once for all    
    parasort(p, a, threads)
    parasort(q, b, threads)
    
    print('Parallel sorting: time =', round(perf_counter() - t0, 3), 'sec\n')

    print("Testing W1, samples of deltas, n=m")
    t0 = perf_counter()
    z = W1(p, q)
    print("Scipy: z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = ot.lp.wasserstein_1d(p, q, p=1)
    print("POT  : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = OT1D(p, q, p=1, sorting=sorting, threads=threads)
    print("SG   : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    print("\nTesting W2, samples of deltas, n=m")
    t0 = perf_counter()
    z = ot.lp.wasserstein_1d(p, q, p=2)
    print("POT  : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = OT1D(p, q, p=2, sorting=sorting, threads=threads)
    print("SG   : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))


    print("\nTesting W1, samples with weights")
    t0 = perf_counter()
    z = W1(p, q, a, b)
    print("Scipy: z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = ot.lp.wasserstein_1d(p, q, a, b, p=1)
    print("POT  : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = OT1D(p, q, a, b, p=1, sorting=sorting, threads=threads)
    print("SG   : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    print("\nTesting W2, samples with weights")
    t0 = perf_counter()
    z = ot.lp.wasserstein_1d(p, q, a, b, p=2)
    print("POT  : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))

    t0 = perf_counter()
    z = OT1D(p, q, a, b, p=2, sorting=sorting, threads=threads)
    print("SG   : z =", round(z, 8), "time =", round(perf_counter() - t0, 4))


def Test3(N = 500000, samples=30, threads=8):
    print('\n--------------- TEST 3: Unsorted input (average runtime) --------------------')

    print('For OT1D using', threads, 'threads\n')

    np.random.seed(13)

    sorting=True

    Ls = [0 for i in range(11)]
    print('\nrunning test', end=' ')
    for _ in range(samples):
        print('.', end=' ', sep='', flush=True)
        
        p = np.random.uniform(1, 10, N)
        q = np.random.uniform(3, 5, N)

        # Random weights for samples
        a = np.random.uniform(500, 1000, N)
        b = np.random.uniform(0, 1000, N)
        a = a/np.sum(a)
        b = b/np.sum(b)    
    
        t0 = perf_counter()
        z = W1(p, q)
        Ls[1] += perf_counter() - t0
    
        t0 = perf_counter()
        z = ot.lp.wasserstein_1d(p, q, p=1)
        Ls[2] += perf_counter() - t0

        t0 = perf_counter()
        z = OT1D(p, q, p=1, sorting=sorting, threads=threads)
        Ls[3] += perf_counter() - t0

        t0 = perf_counter()
        z = ot.lp.wasserstein_1d(p, q, p=2)
        Ls[4] += perf_counter() - t0

        t0 = perf_counter()
        z = OT1D(p, q, p=2, sorting=sorting, threads=threads)
        Ls[5] += perf_counter() - t0

        t0 = perf_counter()
        z = W1(p, q, a, b)
        Ls[6] += perf_counter() - t0
    
        t0 = perf_counter()
        z = ot.lp.wasserstein_1d(p, q, a, b, p=1)
        Ls[7] += perf_counter() - t0
    
        t0 = perf_counter()
        z = OT1D(p, q, a, b, p=1, sorting=sorting, threads=threads)
        Ls[8] += perf_counter() - t0
    
        t0 = perf_counter()
        z = ot.lp.wasserstein_1d(p, q, a, b, p=2)
        Ls[9] += perf_counter() - t0
    
        t0 = perf_counter()
        z = OT1D(p, q, a, b, p=2, sorting=sorting, threads=threads)
        Ls[10] += perf_counter() - t0

    print()

    print("Testing W1, samples of deltas, n=m")
    print("Scipy: average time =", round(Ls[1]/samples, 3),
          "speedup =", round(Ls[1]/Ls[3], 1))
    print("POT  : average time =", round(Ls[2]/samples, 3),
          "speedup =",round(Ls[2]/Ls[3], 1))
    print("SG   : average time =", round(Ls[3]/samples, 3),
          "speedup =",round(Ls[3]/Ls[3], 1))

    print("\nTesting W2, samples of deltas, n=m")
    print("POT  : average time =", round(Ls[4]/samples, 3),
          "speedup =", round(Ls[4]/Ls[5], 1))
    print("SG   : average time =", round(Ls[5]/samples, 3),
          "speedup =", round(Ls[5]/Ls[5], 1))

    print("\nTesting W1, samples with weights")
    print("Scipy: average time =", round(Ls[6]/samples, 3),
          "speedup =", round(Ls[6]/Ls[8], 1))
    print("POT  : average time =", round(Ls[7]/samples, 3),
          "speedup =", round(Ls[7]/Ls[8], 1))
    print("SG   : average time =", round(Ls[8]/samples, 3),
          "speedup =", round(Ls[8]/Ls[8], 1))

    print("\nTesting W2, samples with weights")
    print("POT  : average time =", round(Ls[9]/samples, 3),
          "speedup =", round(Ls[9]/Ls[10], 1))
    print("SG   : average time =", round(Ls[10]/samples, 3),
          "speedup =", round(Ls[10]/Ls[10], 1))


def Test4(N = 500000, samples=30, threads=8):
    print('\n--------------- TEST 4: Sorted input (average runtime) --------------------')

    print('For OT1D using', threads, 'threads\n')

    np.random.seed(13)

    sorting=False

    Ls = [0 for i in range(11)]
    print('running test', end=' ')
    for _ in range(samples):
        print('.', end=' ', sep='', flush=True)
        
        p = np.random.uniform(1, 10, N)
        q = np.random.uniform(3, 5, N)

        # Random weights for samples
        a = np.random.uniform(500, 1000, N)
        b = np.random.uniform(0, 1000, N)
        a = a/np.sum(a)
        b = b/np.sum(b)    

        t0 = perf_counter()
        # Sort input data once for all    
        parasort(p, a, threads)   
        parasort(q, b, threads)
        Ls[0] += perf_counter() - t0
    
        t0 = perf_counter()
        z = W1(p, q)
        Ls[1] += perf_counter() - t0
    
        t0 = perf_counter()
        z = ot.lp.wasserstein_1d(p, q, p=1)
        Ls[2] += perf_counter() - t0

        t0 = perf_counter()
        z = OT1D(p, q, p=1, sorting=sorting, threads=threads)
        Ls[3] += perf_counter() - t0

        t0 = perf_counter()
        z = ot.lp.wasserstein_1d(p, q, p=2)
        Ls[4] += perf_counter() - t0

        t0 = perf_counter()
        z = OT1D(p, q, p=2, sorting=sorting, threads=threads)
        Ls[5] += perf_counter() - t0

        t0 = perf_counter()
        z = W1(p, q, a, b)
        Ls[6] += perf_counter() - t0
    
        t0 = perf_counter()
        z = ot.lp.wasserstein_1d(p, q, a, b, p=1)
        Ls[7] += perf_counter() - t0
    
        t0 = perf_counter()
        z = OT1D(p, q, a, b, p=1, sorting=sorting, threads=threads)
        Ls[8] += perf_counter() - t0
    
        t0 = perf_counter()
        z = ot.lp.wasserstein_1d(p, q, a, b, p=2)
        Ls[9] += perf_counter() - t0
    
        t0 = perf_counter()
        z = OT1D(p, q, a, b, p=2, sorting=sorting, threads=threads)
        Ls[10] += perf_counter() - t0

    print()
    print('Parallel sorting: time =', round(Ls[0]/samples, 3), 'sec\n')

    print("Testing W1, samples of deltas, n=m")
    print("Scipy: average time =", round(Ls[1]/samples, 3),
          "speedup =", round(Ls[1]/Ls[3], 1))
    print("POT  : average time =", round(Ls[2]/samples, 3),
          "speedup =",round(Ls[2]/Ls[3], 1))
    print("SG   : average time =", round(Ls[3]/samples, 3),
          "speedup =",round(Ls[3]/Ls[3], 1))

    print("\nTesting W2, samples of deltas, n=m")
    print("POT  : average time =", round(Ls[4]/samples, 3),
          "speedup =", round(Ls[4]/Ls[5], 1))
    print("SG   : average time =", round(Ls[5]/samples, 3),
          "speedup =", round(Ls[5]/Ls[5], 1))

    print("\nTesting W1, samples with weights")
    print("Scipy: average time =", round(Ls[6]/samples, 3),
          "speedup =", round(Ls[6]/Ls[8], 1))
    print("POT  : average time =", round(Ls[7]/samples, 3),
          "speedup =", round(Ls[7]/Ls[8], 1))
    print("SG   : average time =", round(Ls[8]/samples, 3),
          "speedup =", round(Ls[8]/Ls[8], 1))

    print("\nTesting W2, samples with weights")
    print("POT  : average time =", round(Ls[9]/samples, 3),
          "speedup =", round(Ls[9]/Ls[10], 1))
    print("SG   : average time =", round(Ls[10]/samples, 3),
          "speedup =", round(Ls[10]/Ls[10], 1))

# -----------------------------------------------
#   MAIN function
# -----------------------------------------------
if __name__ == "__main__":
    # Example 1 from here:
    #  https://www.kaggle.com/nhan1212/some-statistical-distances/comments
    
    np.random.seed(13)
    N = 5000000
    
    # Uniform samples
    if False:
        p = np.random.uniform(1, 2, N)
        q = np.random.uniform(0, 1, N)
    else:
        p = np.random.normal(0, 1, N)
        q = np.random.normal(10, 1, N)

    # Random weights for samples
    a = np.random.uniform(500, 1000, N)
    b = np.random.uniform(0, 1000, N)
    a = a/np.sum(a)
    b = b/np.sum(b)    

    Test1(p, q, threads=1)
    Test2(p, q, a, b, threads=1)

    Test3(threads=8)
    Test4(threads=8)